package com.loafle.overflow.proxy;

import com.loafle.overflow.api.OverflowApiServerGrpc;
import com.loafle.overflow.api.ServerInput;
import com.loafle.overflow.api.ServerOutput;
import io.grpc.ServerBuilder;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

/**
 * Created by insanity on 17. 6. 28.
 */
public class ServiceProxy {
    private static final Logger logger = Logger.getLogger(ServiceProxy.class.getName());

    private io.grpc.Server server;
    private ApplicationContext ctx;

    public void start(int port) throws IOException {
        ctx = new AnnotationConfigApplicationContext("com.loafle.overflow");


        server = ServerBuilder.forPort(port)
                .addService(new ServiceImpl(new ServiceInvoker(ctx)))
                .build()
                .start();
        logger.info("Server started, listening on " + port);
        Runtime.getRuntime().addShutdownHook(new Thread() {
            @Override
            public void run() {
                // Use stderr here since the logger may have been reset by its JVM shutdown hook.
                System.err.println("*** shutting down gRPC server since JVM is shutting down");
                ServiceProxy.this.stop();
                System.err.println("*** server shut down");
            }
        });
    }

    public void stop() {
        if (server != null) {
            server.shutdown();
        }
    }

    public void blockUntilShutdown() throws InterruptedException {
        if (server != null) {
            server.awaitTermination();
        }
    }

    static class ServiceImpl extends OverflowApiServerGrpc.OverflowApiServerImplBase {

        private ServiceInvoker serviceInvoker;

        ServiceImpl(ServiceInvoker serviceInvoker) {
            this.serviceInvoker = serviceInvoker;
        }

        @Override
        public void exec(ServerInput request,
                         io.grpc.stub.StreamObserver<ServerOutput> responseObserver) {

            String result = null;
            try {
                result =this.serviceInvoker.invokeForByteString(request.getTarget(), request.getMethod(), request.getParamsList().asByteStringList());

                ServerOutput reply = ServerOutput.newBuilder()
                        .setResult(result)
                        .build();
                responseObserver.onNext(reply);
                responseObserver.onCompleted();
            } catch (NoSuchMethodException e) {
                e.printStackTrace();
                responseObserver.onError(e);
            } catch (ClassNotFoundException e) {
                e.printStackTrace();
                responseObserver.onError(e);
            } catch (IOException e) {
                e.printStackTrace();
                responseObserver.onError(e);
            } catch (InvocationTargetException e) {
                e.printStackTrace();
                responseObserver.onError(e);
            } catch (IllegalAccessException e) {
                e.printStackTrace();
                responseObserver.onError(e);
            }


//            String targetServiceName = request.getTarget();
//            Object service = ctx.getBean(targetServiceName);
//
//            if(service != null) {
//                try {
//
//                    String jsonResult = process(request, service);
//
//                    ServerOutput reply = ServerOutput.newBuilder()
//                            .setResult(jsonResult)
//                            .build();
//                    responseObserver.onNext(reply);
//                    responseObserver.onCompleted();
//                }catch(Exception e) {
//                    e.printStackTrace();
//                    responseObserver.onError(e);
//                }
//
//            }else {
//                responseObserver.onError(new Exception("Not assigned Service :" + service));
//            }
        }

//        private String process(ServerInput request, Object service) throws Exception {
//
//            String methodName = request.getMethod();
//            List<ServerParam> params = request.getParamsList();
//            ObjectMapper mapper = new ObjectMapper();
//            mapper.configure(DeserializationConfig.Feature.FAIL_ON_UNKNOWN_PROPERTIES, false);
//
//            Object retObj = null;
//
//            List<Class> paramTypes = new ArrayList<Class>();
//            List<Object> valueList = new ArrayList<Object>();
//
//            for( ServerParam param : params ){
//
//                if(false == param.getIsCollection()) {
//                    Class<?> cls = Class.forName(param.getType());
//                    Object obj = mapper.readValue(param.getData(), cls);
//                    paramTypes.add(cls);
//                    valueList.add(obj);
//                }else {
//                    String type = param.getType();
//                    int idx = type.indexOf("|");
//                    String firstClassName = type.substring(0, idx);
//                    String lastClassName = type.substring(idx+1);
//                    Class firstCls = Class.forName(firstClassName);
//                    Class<?> lastCls = Class.forName(lastClassName);
//                    mapper.configure(DeserializationConfig.Feature.FAIL_ON_UNKNOWN_PROPERTIES, false);
//                    Object obj = mapper.readValue(param.getData(), mapper.getTypeFactory().constructCollectionType(firstCls, lastCls));
//
//                    paramTypes.add(firstCls);
//                    valueList.add(obj);
//                }
//
//
//            }
//            Method method = null;
//            if(params.size() > 0) {
//                method = MethodSeeker.getMethod(service.getClass(),methodName, paramTypes);
//                if (method == null) {
//                    throw  new Exception("Not found method : " + methodName);
//                }
//                retObj = method.invoke(service, valueList.toArray(new Object[valueList.size()]));
//            }
//            else {
//                method = service.getClass().getMethod(methodName);
//                retObj = method.invoke(service);
//            }
//
//            if(retObj == null) {
//                return "";
//            }
//
//            return mapper.writeValueAsString(retObj);
//        }
    }

}
